// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_BASIC_H
#define OB_BASIC_H

//
// Configuration stuff
//
#include <OB/Config.h>

//
// DLL support
//
#ifdef OB_DLL
#   ifdef OB_BUILD_DLL
#       define OB_IMPORT /**/
#   else
#       define OB_IMPORT __declspec(dllimport)
#   endif
#else
#   define OB_IMPORT /**/
#endif

//
// This is for GNU C++ < 2.8
//
#ifdef __GNUC__ 
#   if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 8)
#       error GNU C++ older than version 2.8 is not supported anymore!
#   endif
#endif

//
// This is for AIX xlC
//
#if defined(_AIX) && !defined(__GNUC__)
#   define HAVE_PRAGMA_DEFINE
#endif

//
// This is for HP aCC
//
#if defined(__hpux)
//
// The definition of "alloca" as "malloc" is necessary since aCC has
// no builtin alloca() and there is also no library that includes an
// alloca() function. Normally substituting alloca() as malloc()
// results in huge memory leaks. However, ORBacus needs alloca() for
// the IDL-to-C++ translator only (for the file idl/Gram.cpp). Since
// the translator terminates and releases all resources immediately
// after doing its job, the leak doesn't matter that much. (Hopefully
// new versions of aCC will have builtin alloca() support or supply a
// library which contains alloca().)
//
#   define alloca(x) malloc(x)
#endif

//
// This is for DEC C++ 6.0-010
//
#ifdef __DECCXX
//
// The DEC C++ compiler has a very strange and annoying exception
// handling bug. In case there is a try block, which does not have a
// local variable with a destructor, the compiler sometimes runs into
// the following internal error:
//
// Internal error: Unexpected null cleanup_block in
// unwind_cleanup_state.IOT/Abort trap (core dumped)
//
// To avoid this, a "dummy" variable with a destructor must be added
// to the try block.
//
#   define OB_DEC_TRY_FIX CORBA::Object_var fixForDECCXX;
#else
#   define OB_DEC_TRY_FIX /**/
#endif

//
// Define mutable keyword as empty if mutable is not supported
//
#ifdef HAVE_NO_MUTABLE
#   define mutable /**/
#endif

//
// Typedef for bool type if bool is not supported
//
#if SIZEOF_BOOL == 0
#   ifndef HAVE_BOOL_TYPEDEF
#       define HAVE_BOOL_TYPEDEF
typedef int bool;
#       ifndef true
const bool true = 1;
#       endif
#       ifndef false
const bool false = 0;
#       endif
#   endif
#endif

//
// Include JThreads/C++ library
//
#ifdef HAVE_JTC

#   ifndef MAKEDEPEND
#       include <JTC/JTC.h>
#   endif


#   define OB_COMMA_MONITOR , public JTCMonitor
#   define OB_COLON_MONITOR : public JTCMonitor
#   define OB_COMMA_RECURSIVE_MUTEX , public JTCRecursiveMutex
#   define OB_COLON_RECURSIVE_MUTEX : public JTCRecursiveMutex
#   define OB_COMMA_MUTEX , public JTCMutex
#   define OB_COLON_MUTEX : public JTCMutex
#   define OB_SYNCHRONIZED JTCSynchronized synchronized(*this);

#else

#   define OB_COMMA_MONITOR /**/
#   define OB_COLON_MONITOR /**/
#   define OB_COMMA_RECURSIVE_MUTEX /**/
#   define OB_COLON_RECURSIVE_MUTEX /**/
#   define OB_COMMA_MUTEX /**/
#   define OB_COLON_MUTEX /**/
#   define OB_SYNCHRONIZED /**/

#endif

//
// Include common header files
//
#include <string.h>
#include <assert.h>

#ifdef HAVE_SYS_TYPES_H
#   include <sys/types.h>
#endif

#ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#endif

#ifdef HAVE_IOSTREAM
#   include <iostream>
#else
#   include <iostream.h>
#endif

//
// In case we have the std:: namespace
//
#ifdef HAVE_STD_IOSTREAM
#   define OB_STD(x) std::x
#else
#   define OB_STD(x) x
#endif

namespace OB
{

//
// Assert macro for "non-compliant application"
//
enum NCAReason
{
    NCANotImpl,         // Not implemented yet
    NCADeprecated,      // Deprecated feature

    NCAWrongAlloc,      // Memory was not allocated with Alloc()
    NCADoubleFree,      // Memory was already freed by Free()

    NCARefCountNotZero, // Object deleted without a reference count of zero
    NCARefCountZero,    // Object already deleted (reference count is zero)

    NCASeqLenGrMax,     // Sequence length greater than maximum length
    NCASeqRange,        // Out of range error in sequence operator[]()
                        // or remove()

    NCANullInit,        // Initialization of T_var type with null pointer
    NCANullPtr,         // operator->() used on null pointer or nil reference
    NCANullDeref,       // Null pointer dereferencing
    NCANullString,      // Null pointer passed as string
    NCANullValue,       // Null value passed as argument

    NCASelfAssign,      // Self assignment causing a dangling pointer

    NCAAnySelfReplace,  // Any content was replaced by it's own value

    NCADiscType,        // Wrong union discriminator type
    NCADiscMismatch,    // Union discriminator mismatch
    NCAUnionInit,       // Uninitialized union used

    NCANoExceptionTC,   // CORBA_Any::operator<<=(Exception*) used for an
                        // exception for which not TypeCode is available

    NCAUnembeddedRecTC, // An operation on an unembedded recursive
                        // TypeCode was invoked
    NCAEmbeddedRecTC    // An already embedded TypeCode was reused
};

void NCA(NCAReason);

#ifdef NDEBUG
 
#    define assert_nca(ex, reason) \
((void)0)
 
#else

#    define assert_nca(ex, reason) \
( ((int)(ex)) ? ((void)0) : (OB::NCA(reason), assert((int)(ex))) )

#   define OB_TRACE 1 // By default if NDEBUG is not defined we will trace
    
#endif

//
// Version and license string
//
OB_IMPORT extern const char* Version;
OB_IMPORT extern const char* License;

//
// Wide string helper functions, platform independent
//
size_t wcslen(const wchar_t*);
wchar_t* wcscpy(wchar_t*, const wchar_t*);
wchar_t* wcscat(wchar_t*, const wchar_t*);
int wcscmp(const wchar_t*, const wchar_t*);

//
// Forward declarations needed by String_var
//
class StrForStruct;
class StrForSeq;

//
// Forward declarations needed for WString_var
//
class WStrForStruct;
class WStrForSeq;

//
// Special "singleton" values for empty strings
// Note: These value must not be const! 
//
OB_IMPORT extern char* NullString;
OB_IMPORT extern wchar_t* NullWString;

} // End of namespace OB

namespace CORBA
{

//
// Basic types
//
#if SIZEOF_CHAR == 1
typedef unsigned char Boolean;
typedef char Char;
typedef unsigned char Octet;
#else
#   error SIZEOF_CHAR must be 1!
#endif

#if SIZEOF_WCHAR_T != 0
typedef wchar_t WChar;
#else
#   error SIZEOF_WCHAR_T must not be 0!
#endif

#if SIZEOF_SHORT == 2
typedef signed short Short;
typedef unsigned short UShort;
#else
#if SIZEOF_INT == 2
typedef signed int Short;
typedef unsigned int UShort;
#else
#   error Neither SIZEOF_SHORT nor SIZEOF_INT is 2!
#endif
#endif

#if SIZEOF_LONG == 4
typedef signed long Long;
typedef unsigned long ULong;
#else
#if SIZEOF_INT == 4
typedef signed int Long;
typedef unsigned int ULong;
#else
#   error Neither SIZEOF_LONG nor SIZEOF_INT is 4!
#endif
#endif

#if SIZEOF_LONG_LONG == 8
#ifndef _MSC_VER
typedef long long LongLong;
typedef unsigned long long ULongLong;
#else
typedef __int64 LongLong;
typedef unsigned __int64 ULongLong;
#endif
#else
#   error SIZEOF_LONG_LONG must be 8!
#endif

#if SIZEOF_FLOAT == 4
typedef float Float;
#else
#   error SIZEOF_FLOAT must be 4!
#endif

#if SIZEOF_DOUBLE == 8
typedef double Double;
#else
#   error SIZEOF_DOUBLE must be 8!
#endif

#if SIZEOF_LONG_DOUBLE >= 12
typedef long double LongDouble;
#else
class LongDouble { };
#undef SIZEOF_LONG_DOUBLE
#define SIZEOF_LONG_DOUBLE 0
#endif

//
// Forward declaration of CORBA::Fixed class
//
class Fixed;

//
// Define _out versions of the basic types
//
typedef Short& Short_out;
typedef Long& Long_out;
typedef LongLong& LongLong_out;
typedef ULongLong& ULongLong_out;
typedef UShort& UShort_out;
typedef ULong& ULong_out;
typedef Float& Float_out;
typedef Double& Double_out;
typedef LongDouble& LongDouble_out;
typedef Char& Char_out;
typedef WChar& WChar_out;
typedef Boolean& Boolean_out;
typedef Octet& Octet_out;

//
// True and false
//
#undef TRUE
#undef FALSE
const Boolean TRUE = 1;
const Boolean FALSE = 0;

//
// CORBA::Flags
//
typedef Long Flags;
const Flags ARG_IN = 1;
const Flags ARG_OUT = 2;
const Flags ARG_INOUT = 3;
const Flags OUT_LIST_MEMORY = 4;
const Flags IN_COPY_VALUE = 5;
const Flags CTX_RESTRICT_SCOPE = 6;

//
// String memory management
//
// - string_alloc/wstring_alloc:
//   Allocates new string (terminating '\0' must not be included in
//   the length argument)
//
// - string_free/wstring_free:
//   Frees a string allocated with string_alloc
//
// - string_dup/wstring_dup:
//   Duplicates a string using the functions string_alloc
//
char* string_alloc(ULong);
wchar_t* wstring_alloc(ULong);

void string_free(char*);
void wstring_free(wchar_t*);

char* string_dup(const char*);
wchar_t* wstring_dup(const wchar_t*);

//
// ORBacus specific:
//
// - string_allocated/wstring_allocated:
//   Returns the length argument that was used with string_alloc
//
// - string_check/wstring_check:
//   Checks if a string was allocated with string_alloc (only
//   works if the library was compiled without NDEBUG set)
//

ULong string_allocated(const char*);
ULong wstring_allocated(const wchar_t*);

#ifdef NDEBUG
inline Boolean string_check(const char*) { return true; }
inline Boolean wstring_check(const wchar_t*) { return true; }
#else
Boolean string_check(const char*);
Boolean wstring_check(const wchar_t*);
#endif

//
// The String_var class
//
class String_var
{
protected:

    char* ptr_;

public:

    //
    // Standard IDL to C++ Mapping
    //
    String_var() : ptr_(0) { }
    String_var(char*);
    String_var(const char* p) : ptr_(string_dup(p)) { }
    String_var(const String_var&);
    String_var(const OB::StrForSeq&);
    String_var(const OB::StrForStruct&);
    ~String_var() { string_free(ptr_); }
    
    String_var& operator=(char*);
    String_var& operator=(const char*);
    String_var& operator=(const String_var&);
    String_var& operator=(const OB::StrForStruct&);
    String_var& operator=(const OB::StrForSeq&);

    const char* in() const { return ptr_; }
    char*& inout() { return ptr_; }
    char*& out();
    char* _retn();

    char& operator[](short idx) { return ptr_[idx]; }
    char operator[](short idx) const { return ptr_[idx]; }
    char& operator[](unsigned short idx) { return ptr_[idx]; }
    char operator[](unsigned short idx) const { return ptr_[idx]; }
    char& operator[](int idx) { return ptr_[idx]; }
    char operator[](int idx) const { return ptr_[idx]; }
    char& operator[](unsigned int idx) { return ptr_[idx]; }
    char operator[](unsigned int idx) const { return ptr_[idx]; }
    char& operator[](long idx) { return ptr_[idx]; }
    char operator[](long idx) const { return ptr_[idx]; }
    char& operator[](unsigned long idx) { return ptr_[idx]; }
    char operator[](unsigned long idx) const { return ptr_[idx]; }

#ifdef HAVE_NO_CONST_TYPE_CONVERSION_OVERLOAD
    operator char*&() const { return (char*&)ptr_; }
#else
#ifndef OB_ONLY_IN_TYPE_CONVERSION
    operator char*&() { return ptr_; }
#endif
    operator const char*() const { return ptr_; }
#endif

    //
    // Additional ORBacus specific functions
    //
    String_var& operator+=(const char*);
    String_var& operator+=(char);
    String_var& operator+=(unsigned char);

    String_var& operator+=(long);
    String_var& operator+=(unsigned long);

    String_var& operator+=(short n)
    { return operator+=((long)n); }
    String_var& operator+=(unsigned short n)
    { return operator+=((unsigned long)n); }

    String_var& operator+=(int n)
    { return operator+=((long)n); }
    String_var& operator+=(unsigned int n)
    { return operator+=((unsigned long)n); }

    //
    // The class String_out needs access to the internals of String_var
    //
    friend class String_out;
};

} // End of namespace CORBA


namespace OB
{

//
// String type used by structs, exceptions and arrays
//
class StrForStruct : public CORBA::String_var
{
public:

    //
    // Standard IDL to C++ Mapping
    //
    StrForStruct() : CORBA::String_var(OB::NullString) { }
    StrForStruct(char* p) : CORBA::String_var(p) { }
    StrForStruct(const char* p) : CORBA::String_var(p) { }
    StrForStruct(const StrForStruct& r) : CORBA::String_var(r) { }
    StrForStruct(const StrForSeq& r) : CORBA::String_var(r) { }

#ifdef WIN32
    StrForStruct& operator=(char* p)
    { String_var::operator=(p); return *this; }
    StrForStruct& operator=(const char* p)
    { String_var::operator=(p); return *this; }
    StrForStruct& operator=(const CORBA::String_var& r)
    { String_var::operator=(r); return *this; }
    StrForStruct& operator=(const StrForStruct& r)
    { String_var::operator=(r); return *this; }
    StrForStruct& operator=(const StrForSeq& r)
    { String_var::operator=(r); return *this; }
#else
    StrForStruct& operator=(char* p)
    { CORBA::String_var::operator=(p); return *this; }
    StrForStruct& operator=(const char* p)
    { CORBA::String_var::operator=(p); return *this; }
    StrForStruct& operator=(const CORBA::String_var& r)
    { CORBA::String_var::operator=(r); return *this; }
    StrForStruct& operator=(const StrForStruct& r)
    { CORBA::String_var::operator=(r); return *this; }
    StrForStruct& operator=(const StrForSeq& r)
    { CORBA::String_var::operator=(r); return *this; }
#endif
};

//
// String type used by string sequences
//
class StrForSeq
{	
    char** ptr_;
    CORBA::Boolean rel_;
    
public:

    //
    // ORBacus specific constructors that work together with the
    // string sequence
    //
    StrForSeq(char** p, CORBA::Boolean rel) : ptr_(p), rel_(rel) { }
    StrForSeq(const StrForSeq& r) : ptr_(r.ptr_), rel_(r.rel_) { }
    
    //
    // Standard IDL to C++ Mapping
    //
    StrForSeq& operator=(char*);
    StrForSeq& operator=(const char*);
    StrForSeq& operator=(const CORBA::String_var&);
    StrForSeq& operator=(const StrForStruct&);
    StrForSeq& operator=(const StrForSeq&);
    
    const char* in() const { return *ptr_; }
    char*& inout() { return *ptr_; }
    char*& out();
    char* _retn();

    char& operator[](short idx) { return (*ptr_)[idx]; }
    char operator[](short idx) const { return (*ptr_)[idx]; }
    char& operator[](unsigned short idx) { return (*ptr_)[idx]; }
    char operator[](unsigned short idx) const { return (*ptr_)[idx]; }
    char& operator[](int idx) { return (*ptr_)[idx]; }
    char operator[](int idx) const { return (*ptr_)[idx]; }
    char& operator[](unsigned int idx) { return (*ptr_)[idx]; }
    char operator[](unsigned int idx) const { return (*ptr_)[idx]; }
    char& operator[](long idx) { return (*ptr_)[idx]; }
    char operator[](long idx) const { return (*ptr_)[idx]; }
    char& operator[](unsigned long idx) { return (*ptr_)[idx]; }
    char operator[](unsigned long idx) const { return (*ptr_)[idx]; }

#ifdef HAVE_NO_CONST_TYPE_CONVERSION_OVERLOAD
    operator char*&() const { return *ptr_; }
#else
#ifndef OB_ONLY_IN_TYPE_CONVERSION
    operator char*&() { return *ptr_; }
#endif
    operator const char*() const { return *ptr_; }
#endif

    friend class CORBA::String_out;
};

} // End of namespace OB


namespace CORBA
{

//
// The String_out class
//
class String_out
{
    // Assignment from String_var and OB::StrForSeq disallowed
    void operator=(const String_var&);
    void operator=(const OB::StrForSeq&);

    char*& ptr_;

public:

    String_out(char*& p)
        : ptr_(p)
    {
	ptr_ = 0;
    }
    
    String_out(String_var& p)
        : ptr_(p.ptr_)
    {
        string_free(ptr_);
        ptr_ = 0;
    }
    
    String_out(OB::StrForSeq& p)
        : ptr_(*p.ptr_)
    {
        string_free(ptr_);
        ptr_ = 0;
    }
    
    String_out(String_out& p)
        : ptr_(p.ptr_)
    {
    }
    
    String_out& operator=(String_out& p)
    {
        ptr_ = p.ptr_;
        return *this;
    }
    
    String_out& operator=(char* p)
    {
        ptr_ = p;
        return *this;
    }
    
    String_out& operator=(const char* p)
    {
        ptr_ = string_dup(p);
        return *this;
    }
    
    operator char*&() { return ptr_; }
    char*& ptr() { return ptr_; }
};

//
// The WString_var class
//
class WString_var
{
protected:

    wchar_t* ptr_;

public:

    //
    // Standard IDL to C++ Mapping
    //
    WString_var() : ptr_(0) { }
    WString_var(wchar_t*);
    WString_var(const wchar_t* p) : ptr_(wstring_dup(p)) { }
    WString_var(const WString_var&);
    WString_var(const OB::WStrForSeq&);
    WString_var(const OB::WStrForStruct&);
    ~WString_var() { wstring_free(ptr_); }
    
    WString_var& operator=(wchar_t*);
    WString_var& operator=(const wchar_t*);
    WString_var& operator=(const WString_var&);
    WString_var& operator=(const OB::WStrForStruct&);
    WString_var& operator=(const OB::WStrForSeq&);

    const wchar_t* in() const { return ptr_; }
    wchar_t*& inout() { return ptr_; }
    wchar_t*& out();
    wchar_t* _retn();

    wchar_t& operator[](short idx) { return ptr_[idx]; }
    wchar_t operator[](short idx) const { return ptr_[idx]; }
    wchar_t& operator[](unsigned short idx) { return ptr_[idx]; }
    wchar_t operator[](unsigned short idx) const { return ptr_[idx]; }
    wchar_t& operator[](int idx) { return ptr_[idx]; }
    wchar_t operator[](int idx) const { return ptr_[idx]; }
    wchar_t& operator[](unsigned int idx) { return ptr_[idx]; }
    wchar_t operator[](unsigned int idx) const { return ptr_[idx]; }
    wchar_t& operator[](long idx) { return ptr_[idx]; }
    wchar_t operator[](long idx) const { return ptr_[idx]; }
    wchar_t& operator[](unsigned long idx) { return ptr_[idx]; }
    wchar_t operator[](unsigned long idx) const { return ptr_[idx]; }

#ifdef HAVE_NO_CONST_TYPE_CONVERSION_OVERLOAD
    operator wchar_t*&() const { return (wchar_t*&)ptr_; }
#else
#ifndef OB_ONLY_IN_TYPE_CONVERSION
    operator wchar_t*&() { return ptr_; }
#endif
    operator const wchar_t*() const { return ptr_; }
#endif

    //
    // Most of these left out because under NT operator+=(unsigned
    // short n) wouldn't work anyway (because operator+=(wchar_t)
    // would be called instead, which is not the same. I don't think
    // these are needed anyway.
    //

    //
    // Additional ORBacus specific functions
    //
    WString_var& operator+=(const wchar_t*);
    WString_var& operator+=(wchar_t);

/*
    WString_var& operator+=(long);
    WString_var& operator+=(unsigned long);

    WString_var& operator+=(short n)
    { return operator+=((long)n); }
#ifndef HAVE_WIN32_WCHAR_TYPE
    WString_var& operator+=(unsigned short n)
    { return operator+=((unsigned long)n); }
#endif
    WString_var& operator+=(int n)
    { return operator+=((long)n); }
    WString_var& operator+=(unsigned int n)
    { return operator+=((unsigned long)n); }
*/

    //
    // The WString_out class needs access to the internals of
    // WString_var
    //
    friend class WString_out;
};

} // End of namespace CORBA


namespace OB
{

//
// WString type used by structs, exceptions and arrays
//
class WStrForStruct : public CORBA::WString_var
{
public:

    //
    // Standard IDL to C++ Mapping
    //
    WStrForStruct() : CORBA::WString_var(OB::NullWString) { }
    WStrForStruct(wchar_t* p) : CORBA::WString_var(p) { }
    WStrForStruct(const wchar_t* p) : CORBA::WString_var(p) { }
    WStrForStruct(const WStrForStruct& r) : CORBA::WString_var(r) { }
    WStrForStruct(const WStrForSeq& r) : CORBA::WString_var(r) { }

#ifdef WIN32
    WStrForStruct& operator=(wchar_t* p)
    { WString_var::operator=(p); return *this; }
    WStrForStruct& operator=(const wchar_t* p)
    { WString_var::operator=(p); return *this; }
    WStrForStruct& operator=(const CORBA::WString_var& r)
    { WString_var::operator=(r); return *this; }
    WStrForStruct& operator=(const WStrForStruct& r)
    { WString_var::operator=(r); return *this; }
    WStrForStruct& operator=(const WStrForSeq& r)
    { WString_var::operator=(r); return *this; }
#else
    WStrForStruct& operator=(wchar_t* p)
    { CORBA::WString_var::operator=(p); return *this; }
    WStrForStruct& operator=(const wchar_t* p)
    { CORBA::WString_var::operator=(p); return *this; }
    WStrForStruct& operator=(const CORBA::WString_var& r)
    { CORBA::WString_var::operator=(r); return *this; }
    WStrForStruct& operator=(const WStrForStruct& r)
    { CORBA::WString_var::operator=(r); return *this; }
    WStrForStruct& operator=(const WStrForSeq& r)
    { CORBA::WString_var::operator=(r); return *this; }
#endif
};

//
// String type used by wstring sequences
//
class WStrForSeq
{	
    wchar_t** ptr_;
    CORBA::Boolean rel_;
    
public:

    //
    // ORBacus specific constructor that works together with
    // the corresponding sequence type
    //
    WStrForSeq(wchar_t** p, CORBA::Boolean rel) : ptr_(p), rel_(rel) { }
    WStrForSeq(const WStrForSeq& r) : ptr_(r.ptr_), rel_(r.rel_) { }
    
    //
    // Standard IDL to C++ Mapping
    //
    WStrForSeq& operator=(wchar_t*);
    WStrForSeq& operator=(const wchar_t*);
    WStrForSeq& operator=(const CORBA::WString_var&);
    WStrForSeq& operator=(const WStrForStruct&);
    WStrForSeq& operator=(const WStrForSeq&);
    
    const wchar_t* in() const { return *ptr_; }
    wchar_t*& inout() { return *ptr_; }
    wchar_t*& out();
    wchar_t* _retn();

    wchar_t& operator[](short idx) { return (*ptr_)[idx]; }
    wchar_t operator[](short idx) const { return (*ptr_)[idx]; }
    wchar_t& operator[](unsigned short idx) { return (*ptr_)[idx]; }
    wchar_t operator[](unsigned short idx) const { return (*ptr_)[idx]; }
    wchar_t& operator[](int idx) { return (*ptr_)[idx]; }
    wchar_t operator[](int idx) const { return (*ptr_)[idx]; }
    wchar_t& operator[](unsigned int idx) { return (*ptr_)[idx]; }
    wchar_t operator[](unsigned int idx) const { return (*ptr_)[idx]; }
    wchar_t& operator[](long idx) { return (*ptr_)[idx]; }
    wchar_t operator[](long idx) const { return (*ptr_)[idx]; }
    wchar_t& operator[](unsigned long idx) { return (*ptr_)[idx]; }
    wchar_t operator[](unsigned long idx) const { return (*ptr_)[idx]; }

#ifdef HAVE_NO_CONST_TYPE_CONVERSION_OVERLOAD
    operator wchar_t*&() const { return *ptr_; }
#else
#ifndef OB_ONLY_IN_TYPE_CONVERSION
    operator wchar_t*&() { return *ptr_; }
#endif
    operator const wchar_t*() const { return *ptr_; }
#endif

    friend class CORBA::WString_out;
};

} // End of namespace OB


namespace CORBA
{

//
// The WString_out class
//
class WString_out
{
    // Assignment from WString_var and OB::WStrForSeq disallowed
    void operator=(const WString_var&);
    void operator=(const OB::WStrForSeq&);

    WChar*& ptr_;

public:

    WString_out(WChar*& p)
        : ptr_(p)
    {
        ptr_ = 0;
    }
    
    WString_out(WString_var& p)
        : ptr_(p.ptr_)
    {
        wstring_free(ptr_);
        ptr_ = 0;
    }

    WString_out(OB::WStrForSeq& p)
        : ptr_(*p.ptr_)
    {
        wstring_free(ptr_);
        ptr_ = 0;
    }
    
    WString_out(WString_out& p)
        : ptr_(p.ptr_)
    {
    }
    
    WString_out& operator=(WString_out& p)
    {
        ptr_ = p.ptr_;
        return *this;
    }
    
    WString_out& operator=(WChar* p)
    {
        ptr_ = p;
        return *this;
    }
    
    WString_out& operator=(const WChar* p)
    {
        ptr_ = wstring_dup(p);
        return *this;
    }
    
    operator WChar*&() { return ptr_; }
    WChar*& ptr() { return ptr_; }
};

//
// Forward declarations needed for OBTypes.h
//
class Any;
class TypeCode;
typedef TypeCode* TypeCode_ptr;

} // End of namespace CORBA

void OBDuplicate(CORBA::TypeCode_ptr);
void OBRelease(CORBA::TypeCode_ptr);

//
// String_var, String_out and OB::StrForSeq output on
// ostream
//
extern OB_STD(ostream)& operator<<(OB_STD(ostream)&, const CORBA::String_var&);
extern OB_STD(ostream)& operator<<(OB_STD(ostream)&, const CORBA::String_out&);
extern OB_STD(ostream)& operator<<(OB_STD(ostream)&, const OB::StrForSeq&);
extern OB_STD(istream)& operator>>(OB_STD(istream)&, CORBA::String_var&);
extern OB_STD(istream)& operator>>(OB_STD(istream)&, CORBA::String_out&);
extern OB_STD(istream)& operator>>(OB_STD(istream)&, OB::StrForSeq&);

//
// WString_var, WString_out and WStrForSeq output on
// OB_STD(ostream)
//
extern OB_STD(ostream)& operator<<(OB_STD(ostream)&,
				   const CORBA::WString_var&);
extern OB_STD(ostream)& operator<<(OB_STD(ostream)&,
				   const CORBA::WString_out&);
extern OB_STD(ostream)& operator<<(OB_STD(ostream)&, const OB::WStrForSeq&);
// TODO: This should be done by a configure script
#if defined (__GNUC__) || defined(__hpux)
#   define HAVE_NO_WCHAR_T_SUPPORT_IN_ISTREAM
#endif
#ifndef HAVE_NO_WCHAR_T_SUPPORT_IN_ISTREAM // See the comment in Basic.cpp
extern OB_STD(istream)& operator>>(OB_STD(istream)&, CORBA::WString_var&);
extern OB_STD(istream)& operator>>(OB_STD(istream)&, CORBA::WString_out&);
extern OB_STD(istream)& operator>>(OB_STD(istream)&, OB::WStrForSeq&);
#endif

namespace OB
{

//
// Byte order constant:
//
// Endian == TRUE (1)  : little endian
// Endian == FALSE (0) : big endian
//
#ifdef WORDS_BIGENDIAN
const CORBA::Boolean Endian = 0;
#else
const CORBA::Boolean Endian = 1;
#endif

//
// Memory allocation functions
//
// - Alloc:
//   Allocates memory
//
// - Free:
//   Frees memory allocated with Alloc
//
// - Allocated:
//   Returns the unsigned long argument that was used with Alloc
//
// - AllocCheck:
//   Checks if memory was allocated with Alloc (does only
//   work if the library was compiled without NDEBUG set)
//
void* Alloc(size_t, CORBA::ULong);
void Free(void*);
CORBA::ULong Allocated(void*);
#ifdef NDEBUG
inline bool AllocCheck(const void*) { return true; }
#else
bool AllocCheck(const void*);
#endif

//
// Memory allocation functions used by string sequences
//
char** StrSeqAllocbuf(CORBA::ULong);
void StrSeqFreebuf(char**);
wchar_t** WStrSeqAllocbuf(CORBA::ULong);
void WStrSeqFreebuf(wchar_t**);

//
// Convert between octet sequence and ascii string
//
char* OctetsToAscii(const CORBA::Octet*, CORBA::ULong);
CORBA::Octet* AsciiToOctets(const char*);

//
// "Simple" reference counting class, not thread-safe
//
class SimpleRefCount
{
    CORBA::Long ref_; // The reference count

    //
    // Hide copy-constructor and assignment operator
    //
    SimpleRefCount(const SimpleRefCount&);
    void operator=(const SimpleRefCount&);

public:

    SimpleRefCount() : ref_(1) { }
    virtual ~SimpleRefCount()
    { assert_nca(ref_ == 0, NCARefCountNotZero); }

    //
    // Increment reference count
    //
    void _OB_incRef() { ++ref_; }

    //
    // Decrement reference count - delete object if reference count
    // falls to zero
    //
    void _OB_decRef()
    {
	assert_nca(ref_ > 0, NCARefCountZero);
	if(--ref_ == 0)
	    delete this;
    }

    //
    // Get and set reference count
    //
    CORBA::ULong _OB_getRef() const { return ref_; }
    void _OB_setRef(CORBA::ULong ref) { ref_ = ref; }
};

//
// Reference counting class, thread-safe
//
#ifndef HAVE_JTC

typedef SimpleRefCount RefCount;

#else

#ifdef WIN32

class RefCount
{
    CORBA::Long ref_; // The reference count

    //
    // Hide copy-constructor and assignment operator
    //
    RefCount(const RefCount&);
    void operator=(const RefCount&);

public:

    RefCount() : ref_(1) { }
    virtual ~RefCount()
    { assert_nca(ref_ == 0, NCARefCountNotZero); }

    //
    // Increment reference count
    //
    void _OB_incRef() { InterlockedIncrement(&ref_); }

    //
    // Decrement reference count - delete object if reference count
    // falls to zero
    //
    void _OB_decRef()
    {
	assert_nca(ref_ > 0, NCARefCountZero);
	if(InterlockedDecrement(&ref_) == 0)
	    delete this;
    }

    //
    // Get and set reference count
    //
    CORBA::ULong _OB_getRef() const { return ref_; }
    void _OB_setRef(CORBA::ULong ref) { ref_ = ref; }
};

#else // !WIN32

class RefCount
{
    CORBA::Long ref_; // The reference count
    JTCMutex mut_; // The reference count mutex

    //
    // Hide copy-constructor and assignment operator
    //
    RefCount(const RefCount&);
    void operator=(const RefCount&);

public:

    RefCount() : ref_(1) { }
    virtual ~RefCount() { assert_nca(ref_ == 0, NCARefCountNotZero); }

    //
    // Increment reference count
    //
    void _OB_incRef() { JTCSynchronized sync(mut_); ++ref_; }

    //
    // Decrement reference count - delete object if reference count
    // falls to zero
    //
    void _OB_decRef()
    {
	bool deleteThis;
	
	{
	    JTCSynchronized sync(mut_);
	    assert_nca(ref_ > 0, NCARefCountZero);
	    deleteThis = (--ref_ == 0);
	}

	//
	// Pretty awkward to delete an object like this but a bug in
	// the Sun CC 4.2 forces us to do it this way.
	//
	if(deleteThis)
	    delete this;
    }

    //
    // Get and set reference count
    //
    CORBA::ULong _OB_getRef() const { return ref_; }
    void _OB_setRef(CORBA::ULong ref)
    { JTCSynchronized sync(mut_); ref_ = ref; }
};

#endif // WIN32

#endif // HAVE_JTC

//
// A simple buffer struct
//

#ifdef WIN32

//
// Under Win'95 (and possibly '98) the allocated buffer is not
// necessarily 8 byte aligned. This does not occur under NT. Therefore
// this special buffer class always ensures this.
//
struct Buffer
{
    CORBA::Octet* real; // Points to the "real" buffer

    //
    // Hide default constructor, copy-constructor and assignment
    // operator
    //
    Buffer()
	: real(0), data(0), len(0)
    {
    }

    Buffer(const Buffer& buf)
	: real(new CORBA::Octet[buf.len + 7]), len(buf.len)
    {
	data = real + 7;
	data -= (unsigned int)data % 8;
	memcpy(data, buf.data, len);
    }

    Buffer& operator=(const Buffer& buf)
    {
	if(&buf != this)
	{
	    delete [] real;
	    len = buf.len;
	    real = new CORBA::Octet[len + 7];
	    data = real + 7;
	    data -= (unsigned int)data % 8;
	    memcpy(data, buf.data, len);
	}
	return *this;
    }
    
public:

    Buffer(CORBA::ULong n)
	: real(new CORBA::Octet[n + 7]), len(n)
    {
	data = real + 7;
	data -= (unsigned int)data % 8;
#ifdef OB_CLEAR_MEM
	memset(data, 0, len);
#endif
    }

    ~Buffer()
    {
	delete [] real;
    }

    void alloc(CORBA::ULong n)
    {
	delete [] real;
	len = n;
	real = new CORBA::Octet[len + 7];
	data = real + 7;
	data -= (unsigned int)data % 8;
#ifdef OB_CLEAR_MEM
	memset(data, 0, n);
#endif
    }

    void realloc(CORBA::ULong n)
    {
        //
	// Shrinking the buffer is illegal
        //
	assert(n >= len);

	if(n > len)
	{
	    CORBA::Octet* newReal = new CORBA::Octet[n + 7];
	    CORBA::Octet* newData = newReal + 7;
	    newData -= (unsigned int)newData % 8;
	    memcpy(newData, data, len);
#ifdef OB_CLEAR_MEM
	    memset(newData + len, 0, n - len);
#endif
	    delete [] real;
	    real = newReal;
	    data = newData;
	    len = n;
	}
    }

    void consume(Buffer& buf)
    {
	delete [] real;
	real = buf.real;
	data = buf.data;
	len = buf.len;
	buf.real = 0;
	buf.data = 0;
	buf.len = 0;
    }

    CORBA::Octet* data; // Points to the buffer
    CORBA::ULong len; // The buffer length
};    

#else

struct Buffer
{
    //
    // Hide default constructor, copy-constructor and assignment
    // operator
    //
    Buffer()
	: data(0), len(0)
    {
    }

    Buffer(const Buffer& buf)
	: data(new CORBA::Octet(buf.len)), len(buf.len)
    {
	memcpy(data, buf.data, len);
    }

    Buffer& operator=(const Buffer& buf)
    {
	if(&buf != this)
	{
	    delete [] data;
	    len = buf.len;
	    data = new CORBA::Octet[len];
	    memcpy(data, buf.data, len);
	}
	return *this;
    }

public:

    Buffer(CORBA::ULong n)
	: data(new CORBA::Octet[n]), len(n)
    {
#ifdef OB_CLEAR_MEM
	memset(data, 0, len);
#endif
    }

    ~Buffer()
    {
	delete [] data;
    }

public:

    void alloc(CORBA::ULong n)
    {
	delete [] data;
	data = new CORBA::Octet[n];
	len = n;
#ifdef OB_CLEAR_MEM
	memset(data, 0, n);
#endif
    }

    void realloc(CORBA::ULong n)
    {
	// Shrinking the buffer is illegal
	assert(n >= len);

	if(n > len)
	{
	    CORBA::Octet* newData = new CORBA::Octet[n];
	    memcpy(newData, data, len);
#ifdef OB_CLEAR_MEM
	    memset(newData + len, 0, n - len);
#endif
	    delete [] data;
	    data = newData;
	    len = n;
	}
    }

    void consume(Buffer& buf)
    {
	delete [] data;
	data = buf.data;
	len = buf.len;
	buf.data = 0;
	buf.len = 0;
    }

    CORBA::Octet* data; // Points to the buffer
    CORBA::ULong len; // The buffer length
};

#endif

//
// Forward declarations for marshalling stream classes
//
class InputStream;
class OutputStream;
class CountStream;

} // End of namespace OB

#endif
